# -*- coding: utf-8 -*-
# avbroadcast - republish media streams for mass consumption
# (c) 2018-2019 Andreas Motl <andreas.motl@elmyra.de>
import logging
import os
import sys

from docopt import docopt
from avbroadcast import __version__

from avbroadcast.core import InputStream, RtmpHlsPipeline
from avbroadcast.util import make_progress_filename, normalize_options, boot_logging
from avbroadcast.watch import watch_filesystem

logger = logging.getLogger(__name__)

APP_NAME = 'avbroadcast'


def run():
    """
    Usage:
        {program} ingest --stream=<stream> [--base-port=<base-port>] [--verbose]
        {program} publish --name=<name> [--base-port=<base-port>] --target=<target> [--verbose]
        {program} io --name=<name> --stream=<stream> --target=<target> [--base-port=<base-port>] [--verbose] [--tmux] [--analyze]
        {program} watch --path=<path>
        {program} info
        {program} --version
        {program} (-h | --help)

    Options:
        --base-port=<base-port>     Use this port as a baseline for forwarding the first UDP stream.
                                    [default: 50000]
        --verbose                   Increase verbosity

    Examples:

        # Ingest media stream from RTMP.
        avbroadcast ingest --stream="rtmp://184.72.239.149/vod/mp4:bigbuckbunny_450.mp4?reuse=1" --base-port=50000

        # Package using HLS and publish to local file system.
        avbroadcast publish --name="bigbuckbunny" --base-port=50000 --target="/var/spool/hls-local"

        # Watch output directory
        avbroadcast watch --path=/var/spool/hls-local

        # Package using HLS and publish to HTTP server.
        avbroadcast publish --name="bigbuckbunny" --base-port=50000 --target="http://localhost:6767/hls-live"

    """

    # Use generic commandline options schema and amend with current program name.
    commandline_schema = run.__doc__.format(program=APP_NAME)

    d = docopt(commandline_schema, version=APP_NAME + ' ' + __version__)

    # Read commandline options.
    options = normalize_options(docopt(commandline_schema, version=APP_NAME + ' ' + __version__))

    # Start logging subsystem.
    if options['verbose']:
        options['debug'] = True
    boot_logging(options)

    # Report about runtime options.
    logger.info(options)

    # tmux new -s toptop 'htop' \; split-window 'iotop' \; select-layout even-horizontal
    # tmux attach -t avb \; split-window htop --delay=3 \; split-window -h glances --time=0.3
    # select-pane -t0
    if options['tmux']:
        real = list(sys.argv)
        real.remove('--tmux')
        real_command = ' '.join(real)
        #print(real_command)
        #tmux_command = "tmux -2 new-session -s avb \; new-window -t avb '{}'".format(real_command)
        tmux_command = "tmux new-session -s avb '{}'".format(real_command)

        if options['analyze']:
            tmux_command += ' \; split-window htop --delay=3 \; split-window -h glances --time=0.3'
            tmux_command += ' \; select-pane -t0 \; split-window -h avbroadcast watch --path=/var/spool/hls-local'

        print(tmux_command)
        os.system(tmux_command)
        return

    # Dispatch to core methods.
    pipeline = RtmpHlsPipeline()
    if options['ingest']:
        pipeline.ingest(options['stream'], int(options['base-port']))

    if options['publish']:
        pipeline.publish(options['name'], int(options['base-port']), options['target'])

    if options['io']:
        pipeline.ingest(options['stream'], int(options['base-port']))
        pipeline.publish(options['name'], int(options['base-port']), options['target'])

    if options['watch']:
        # TODO: Propagate resolutions and interval
        watch_filesystem(options['path'])

    #if options['add-watch']:
    #    # TODO: Propagate resolutions and interval
    #    watch_filesystem(options['target'], clear_screen=False)



# -*- coding: utf-8 -*-
# (c) 2018-2019 Andreas Motl <andreas.motl@elmyra.de>
import logging
import os
import re
import sys
import getopt
import textwrap
import tempfile
import unicodedata


def boot_logging(options=None):
    log_level = logging.INFO
    if options and options.get('debug'):
        log_level = logging.DEBUG
    setup_logging(level=log_level)


def setup_logging(level=logging.INFO):
    log_format = '%(asctime)-15s [%(name)-20s] %(levelname)-7s: %(message)s'
    logging.basicConfig(
        format=log_format,
        stream=sys.stderr,
        level=level)


def normalize_options(options):
    normalized = {}
    for key, value in options.items():
        key = key.strip('--<>')
        normalized[key] = value
    return normalized


def sanitize_text(text):
    text = textwrap.dedent(text)
    text = text.strip()
    return text


def slugify(value):
    """
    Slugify function from django.utils.text, slightly modified.

    Converts to lowercase, removes non-word characters (alphanumerics and
    underscores) and converts spaces to hyphens. Also strips leading and
    trailing whitespace.

    https://stackoverflow.com/questions/5574042/string-slugification-in-python/27264385#27264385
    """
    value = unicodedata.normalize('NFKD', value).encode('ascii', 'ignore').decode('ascii')
    value = re.sub('[^\w\s-]', '-', value).strip().lower()
    value = re.sub('[-\s]+', '-', value)
    return value


def make_progress_filename(uri):
    dirname = tempfile.gettempdir()
    filename = slugify(uri) + '.ffmpeg.progress'
    filepath = os.path.join(dirname, filename)
    return filepath



# -*- coding: utf-8 -*-
# avbroadcast - republish media streams for mass consumption
# (c) 2019 Andreas Motl <andreas.motl@elmyra.de>
import os
import time
import logging

log = logging.getLogger(__name__)

APP_NAME = 'avbroadcast'


def watch_filesystem(path, resolutions=None, clear_screen=True, interval=0.5):
    while True:
        if clear_screen:
            os.system('clear')
        log.info('Scanning {path} each {interval} seconds.'.format(path=path, interval=interval))

        scan_filesystem(path, resolutions=resolutions)
        time.sleep(interval)


def scan_filesystem(path, resolutions=None):
    resolutions = resolutions or "144 240 360 480 720 1080"
    for resolution in resolutions.split(' '):
        command = 'ls {path} | grep "\-{resolution}\-" | sort | tail -n1'.format(path=path, resolution=resolution)
        os.system(command)



# -*- coding: utf-8 -*-
import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.rst')).read()

setup(name='avbroadcast',
      version='0.5.0',
      description='avbroadcast - republish media streams for mass consumption',
      long_description=README,
      license="AGPL 3, EUPL 1.2",
      classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "License :: OSI Approved :: European Union Public Licence 1.2 (EUPL 1.2)",
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Manufacturing",
        "Intended Audience :: Science/Research",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Telecommunications Industry",
        "Topic :: Communications",
        "Topic :: Internet",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: HTTP Servers",
        "Topic :: Multimedia :: Sound/Audio",
        "Topic :: Multimedia :: Sound/Audio :: Conversion",
        "Topic :: Multimedia :: Video",
        "Topic :: Multimedia :: Video :: Conversion",
        "Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator",
        "Topic :: Software Development :: Libraries",
        "Topic :: System :: Archiving",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Operating System :: MacOS"
      ],
      author='Andreas Motl',
      author_email='andreas.motl@elmyra.de',
      url='https://github.com/daq-tools/avbroadcast',
      keywords='audio video broadcast rtmp hls http ' +
               '',
      packages=find_packages(),
      include_package_data=True,
      package_data={
      },
      zip_safe=False,
      test_suite='avbroadcast.test',
      install_requires=[
          'docopt==0.6.2',
      ],
      dependency_links=[
      ],
      entry_points={
          'console_scripts': [
              'avbroadcast = avbroadcast.commands:run',
          ],
      },
)



# -*- coding: utf-8 -*-
# avbroadcast - republish media streams for mass consumption
# (c) 2018-2019 Andreas Motl <andreas.motl@elmyra.de>
import shlex
import logging
import subprocess
from threading import Thread

from avbroadcast.util import sanitize_text, make_progress_filename

logger = logging.getLogger(__name__)


class RtmpHlsPipeline:

    presets = [
        {
            'resolution': 1080,
            'ffmpeg_bandwidth': '-b:v 4000k -maxrate 6000k -bufsize 4800k',
            'address': 'udp://127.0.0.1:{port}',
        },
        {
            'resolution': 240,
            'ffmpeg_bandwidth': '-b:v  450k -maxrate  675k -bufsize  540k',
            'address': 'udp://127.0.0.1:{port}',
        },
        {
            'resolution': 360,
            'ffmpeg_bandwidth': '-b:v  990k -maxrate 1485k -bufsize 1188k',
            'address': 'udp://127.0.0.1:{port}',
        },
        {
            'resolution': 480,
            'ffmpeg_bandwidth': '-b:v 1600k -maxrate 2400k -bufsize 1920k',
            'address': 'udp://127.0.0.1:{port}',
        },
        {
            'resolution': 720,
            'ffmpeg_bandwidth': '-b:v 2000k -maxrate 3000k -bufsize 2400k',
            'address': 'udp://127.0.0.1:{port}',
        },
    ]

    def stream_descriptions(self, base_port):
        for preset in self.presets:

            stream = preset.copy()
            port = base_port + preset['resolution']
            stream['address'] = stream['address'].format(port=port)

            yield stream

    def ingest(self, stream, base_port):
        decoder = InputStream(source=stream, base_port=base_port)
        command = decoder.configure(presets=self.stream_descriptions(base_port))

        pc = PipelineCommand(command)
        pc.start()

    def publish(self, name, base_port, upload_url):
        packager = OutputPackager(name=name, base_port=base_port, upload_url=upload_url)
        command = packager.configure(presets=self.stream_descriptions(base_port))

        pc = PipelineCommand(command)
        pc.start()


class PipelineCommand(Thread):

    def __init__(self, command):
        super().__init__()
        self.command = command

    def run(self):
        cmd = list(map(str.strip, shlex.split(self.command)))
        #logger.debug(cmd)
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE)

        with process as proc:
            logger.write(proc.stdout.read())


class InputStream:

    # TODO: -report

    ffmpeg_base = """
        ffmpeg \\
        \\
        -fflags nobuffer \\
        -threads 0 -y -progress "{progress_file}" \\
        -i "{source}" \\
        \\
        -acodec aac -ac 2 -ar 48000 -b:a 128k
    """

    # {resolution}p video is streamed to {address}
    ffmpeg_video_stream = """
        -vf scale="trunc(oh*a/2)*2:{resolution}" \\
        -map_metadata -1 -pix_fmt yuv420p -vcodec libx264 -preset:v superfast \\
        -force_key_frames "expr:gte(t,n_forced*2)" \\
        -x264opts ref=1:no-cabac=1:bframes=0:b-pyramid=0:scenecut=0 -sc_threshold 0 \\
        -profile:v main {ffmpeg_bandwidth} -level 31 \\
        -filter:a "pan=stereo|c0=c0|c1=c0" \\
        -f mpegts "{address}" -map 0:v
    """

    def __init__(self, source=None, base_port=None):
        self.source = source
        self.base_port = base_port
        self.progress_file = make_progress_filename(source)

    def get_command(self, presets):

        ffmpeg_base = sanitize_text(self.ffmpeg_base)
        data = {}
        data['source'] = self.source
        data['progress_file'] = self.progress_file
        ffmpeg_base = ffmpeg_base.format(**data)
        parts = [ffmpeg_base]

        for preset in presets:
            stream_part = sanitize_text(self.ffmpeg_video_stream).format(**preset)
            parts.append(stream_part)

        command = ' \\\n\\\n'.join(parts)
        return command

    def configure(self, presets):
        command = self.get_command(presets)
        logger.debug("Running ffmpeg command:\n%s", command)

        #command_suffix = '1> {logfile} 2>&1'.format(logfile='/$TMPDIR/avbroadcast_ffmpeg_6028_2019-03-18_19-08-05.log')
        #command += command_suffix

        return command

        #command_suffix += ' &'
        #os.system(command)
        #os.execlp(command)

        #element = PipelineElement(command)
        #print(element)


class OutputPackager:

    packager_base = """
        packager \\
        --io_block_size 65536 --fragment_duration 2 --segment_duration 2 \\
        --time_shift_buffer_depth 3600 --preserved_segments_outside_live_window 7200 \\
        --hls_master_playlist_output "{upload_url}/{name}.m3u8" \\
        --hls_playlist_type LIVE \\
        --vmodule=file=1,http_file=1,media_playlist=1,hls_notify_muxer_listener=1,ts_segmenter=1,ts_writer=1
    """

    # TODO: Optionally add option "--libcurl_verbosity=3"
    # TODO: Optionally add "--vmodule" options "buffer_writer=1", "master_playlist=1" and "packed_audio_writer=1"

    packager_audio_stream = """
        "input={address}?reuse=1,stream=audio,segment_template={upload_url}/{name}-audio-$Number%04d$.aac,playlist_name={name}-audio.m3u8,hls_group_id=audio"
    """

    packager_video_stream = """
        "input={address}?reuse=1,stream=video,segment_template={upload_url}/{name}-video-{resolution}-$Number%04d$.ts,playlist_name={name}-video-{resolution}.m3u8"
    """

    newline_token = ' \\\n'

    def __init__(self, name=None, base_port=None, upload_url=None):
        self.name = name
        self.base_port = base_port
        self.upload_url = upload_url

    def get_command(self, presets):
        packager_base = sanitize_text(self.packager_base)

        data = {}
        data['name'] = self.name
        data['upload_url'] = self.upload_url
        packager_base = packager_base.format(**data)

        presets = list(presets)
        stream_parts = []

        # 1. Add audio stream from first channel
        first_preset = presets[0]
        stream_part = sanitize_text(self.packager_audio_stream).format(**data, **first_preset)
        stream_parts.append(stream_part)

        # 2. Add all video stream channels
        for preset in presets:
            stream_part = sanitize_text(self.packager_video_stream).format(**data, **preset)
            stream_parts.append(stream_part)

        all_streams = self.join_command(stream_parts)

        parts = [packager_base, all_streams]
        command = self.join_command(parts)
        return command

    def join_command(self, parts):
        return self.newline_token.join(parts)

    def configure(self, presets):
        command = self.get_command(presets)
        logger.debug("Running packager command:\n%s", command)

        #command_suffix = '1> {logfile} 2>&1'.format(logfile='/$TMPDIR/avbroadcast_packager_6028_2019-03-18_19-08-05.log')

        #os.system(command)
        return command
